package ctbrec.recorder.download.hls;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.nio.file.Files;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.iheartradio.m3u8.ParseException;
import com.iheartradio.m3u8.PlaylistException;

import ctbrec.Config;
import ctbrec.Hmac;
import ctbrec.Model;
import ctbrec.OS;
import ctbrec.io.HttpClient;
import ctbrec.io.StreamRedirectThread;
import ctbrec.recorder.ProgressListener;
import ctbrec.recorder.RecordingManager;
import okhttp3.Request;
import okhttp3.Response;

public class MergedHlsDownload extends HlsDownload {
    private static final Logger LOG = LoggerFactory.getLogger(MergedHlsDownload.class);

    private File finalFile;

    public MergedHlsDownload(HttpClient client) {
        super(client);
    }

    @Override
    public void init(Config config, Model model) {
        super.init(config, model);
        finalFile = Config.getInstance().getFileForRecording(model, "mp4");
    }

    @Override
    public void postprocess(ctbrec.Recording recording) {
        super.postprocess(recording);
        File dir = new File(Config.getInstance().getSettings().recordingsDir, recording.getPath());
        File playlist = new File(dir, "playlist.m3u8");
        if (!playlist.exists()) {
            super.generatePlaylist(recording);
        }

        try {
            postprocess(playlist, finalFile);
            String recordingsDir = Config.getInstance().getSettings().recordingsDir;
            String path = finalFile.getAbsolutePath().substring(recordingsDir.length());
            recording.setPath(path);
            if (dir.list().length == 0) {
                RecordingManager.deleteEmptyParents(dir);
            }
        } catch (PostProcessingException | IOException e) {
            LOG.error("An error occurred during post-processing", e);
        }
    }

    private void postprocess(File playlist, File target) throws PostProcessingException {
        try {
            File dir = playlist.getParentFile();
            // @formatter:off
            String[] cmdline = OS.getFFmpegCommand(
                    "-i", playlist.getAbsolutePath(),
                    "-c:v", "copy",
                    "-c:a", "copy",
                    "-movflags", "faststart",
                    "-f", "mp4",
                    target.getAbsolutePath()
                    );
            // @formatter:on
            LOG.debug("Command line: {}", Arrays.toString(cmdline));
            Process ffmpeg = Runtime.getRuntime().exec(cmdline, new String[0], playlist.getParentFile());
            new Thread(new StreamRedirectThread(ffmpeg.getInputStream(), System.out)).start(); // NOSONAR
            new Thread(new StreamRedirectThread(ffmpeg.getErrorStream(), System.err)).start(); // NOSONAR
            int exitCode = ffmpeg.waitFor();
            if (exitCode == 0) {
                Files.delete(playlist.toPath());
                File[] segments = dir.listFiles((directory, filename) -> filename.endsWith(".ts"));
                for (File segment : segments) {
                    Files.delete(segment.toPath());
                }
            } else {
                throw new PostProcessingException("FFmpeg exit code was " + exitCode);
            }
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            LOG.error("Interrupted while waiting for FFMPEG", e);
        } catch (IOException e) {
            LOG.error("Couldn't execute FFMPEG", e);
        }
    }

    public void downloadFinishedRecording(String segmentPlaylistUri, File target, ProgressListener progressListener)
            throws IOException, ParseException, PlaylistException, InvalidKeyException, NoSuchAlgorithmException, PostProcessingException {
        if (Config.getInstance().getSettings().requireAuthentication) {
            URL u = new URL(segmentPlaylistUri);
            String path = u.getPath();
            byte[] key = Config.getInstance().getSettings().key;
            if (!Config.getInstance().getContextPath().isEmpty()) {
                path = path.substring(Config.getInstance().getContextPath().length());
            }
            String hmac = Hmac.calculate(path, key);
            segmentPlaylistUri = segmentPlaylistUri + "?hmac=" + hmac;
        }

        File tempDir = new File(target.getParentFile(), "ctbrec-download-tmp-" + target.getName());
        Files.createDirectories(tempDir.toPath());

        downloadFile(segmentPlaylistUri, tempDir);
        SegmentPlaylist segmentPlaylist = getNextSegments(segmentPlaylistUri);
        int fileCounter = 0;
        for (String segmentUrl : segmentPlaylist.segments) {
            downloadFile(segmentUrl, tempDir);
            fileCounter++;
            int total = segmentPlaylist.segments.size();
            int progress = (int) (fileCounter / (double) total * 100);
            progressListener.update(progress);
        }

        File downloadedPlaylist = new File(tempDir, "playlist.m3u8");
        postprocess(downloadedPlaylist, target);
        Files.delete(tempDir.toPath());
    }

    private void downloadFile(String fileUri, File tempDir) throws IOException {
        LOG.trace("Downloading file {} to {}", fileUri, tempDir);
        Request request = new Request.Builder().url(fileUri).addHeader("connection", "keep-alive").build();
        Response response = client.execute(request);
        InputStream in = null;
        File file = new File(request.url().encodedPath());
        try (FileOutputStream fos = new FileOutputStream(new File(tempDir, file.getName()))) {
            in = response.body().byteStream();
            byte[] b = new byte[1024 * 100];
            int length = -1;
            while ((length = in.read(b)) >= 0) {
                fos.write(b, 0, length);
            }
        } finally {
            if (in != null) {
                in.close();
            }
            response.close();
        }
    }
}
